from pyfileconf.basemodels.file import ConfigFileBase
from pyfileconf.io.file.interfaces.datadict import SpecificClassDictInterface

class SpecificClassDictFile(ConfigFileBase):
    # lines to always import. pass import objects
    # no need to override default
    # always_imports = []

    # assignment lines to always include at beginning. pass assign objects
    # no need to override default
    # always_assigns = []

    # class to use for interfacing with file
    # no need to override default
    interface_class = SpecificClassDictInterface

    def load(self) -> dict:  # type: ignore
        return self.interface.load()


from copy import deepcopy

from pyfileconf import Selector, PipelineManager
from pyfileconf.iterate import IterativeRunner
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf import context
from tests.input_files.amodule import SecondExampleClass, a_function
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST, SAME_CLASS_CONFIG_DICT_LIST, \
    DIFFERENT_CLASS_CONFIG_DICT_LIST


class TestPipelineManagerRun(PipelineManagerTestBase):

    def test_run_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)

    def test_run_function_multiple(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run([iv, iv])
        assert result == [(None, None), (None, None)]

    def test_run_function_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 run
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)

        # Assert pipeline manager 2 run
        iv = sel.test_pipeline_manager2.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)

    def test_create_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(None)

    def test_run_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        result = pipeline_manager.run(iv)
        assert result == 'woo'

    def test_create_class_multiple_pms(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 create
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(None)

        # Assert pipeline manager 2 create
        ec = sel.test_pipeline_manager2.stuff.ExampleClass()
        assert ec == ExampleClass(None)

    def test_create_from_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_run_from_specific_class_dict_original_execute_key(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        result = pipeline_manager.run(iv)
        assert result == 'woo'

    def test_run_from_specific_class_dict_custom_execute_key(self):
        self.write_example_class_dict_to_file()
        ccdl = deepcopy(CLASS_CONFIG_DICT_LIST)
        for config_dict in ccdl:
            config_dict['execute_attr'] = 'my_call'
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=ccdl
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        result = pipeline_manager.run(iv)
        result2 = iv()
        assert result == result2 == 'woo2'

    def test_get_dependencies_while_running(self):
        self.write_example_class_dict_to_file()
        ccdl = deepcopy(CLASS_CONFIG_DICT_LIST)
        for config_dict in ccdl:
            config_dict['execute_attr'] = 'dependent_call'
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=ccdl
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.stuff.data2')
        pipeline_manager.create('example_class.stuff.data3')
        pipeline_manager.create('ec', ExampleClass)
        pipeline_manager.create('ec2', ExampleClass)
        pipeline_manager.create('ec3', ExampleClass)
        pipeline_manager.create('af', a_function)
        pipeline_manager.create('af2', a_function)
        pipeline_manager.create('af3', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        result = pipeline_manager.run(iv)
        assert context.config_dependencies == context.force_update_dependencies \
               == self.expect_pm_1_specific_class_depends_on_pm_1_specific_class_3_class_1_2_function_1_2

    def test_create_from_specific_class_dict_multiple_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 create
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

        # Assert pipeline manager 2 create
        ec = sel.test_pipeline_manager2.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_create_from_multiple_specific_class_dicts_same_class(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(1)  # example_class2
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=SAME_CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        ec2 = sel.test_pipeline_manager.example_class2.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == ec2.name == expect_ec.name
        assert ec.a == ec2.a == expect_ec.a

    def test_create_from_multiple_specific_class_dicts_different_class(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(2)  # second_example_class
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=DIFFERENT_CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        sec = sel.test_pipeline_manager.second_example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        expect_sec = SecondExampleClass(name='data')
        assert ec.name == expect_sec.name == expect_ec.name
        assert ec.a == expect_ec.a
        assert sec.b == expect_sec.b


class TestPipelineManagerRunIter(PipelineManagerTestBase):

    def test_run_iter_function_single(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        cd = dict(
            section_path_str='stuff.a_function',
            b=10
        )
        config_dicts = [cd]
        result = pipeline_manager.run_iter(iv, config_dicts)
        assert result == [((cd,), (None, 10))]
        result = pipeline_manager.run_iter(iv, config_dicts, collect_results=False)
        assert result == []

    def test_run_iter_function_multiple(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        cd = dict(
            section_path_str='stuff.a_function',
            b=10,
            a=2
        )
        cd2 = dict(
            section_path_str='stuff.a_function',
            b=20
        )
        config_dicts = [cd, cd2]
        result = pipeline_manager.run_iter(iv, config_dicts)
        assert result == [((cd,), (2, 10)), ((cd2,), (None, 20))]
        result = pipeline_manager.run_iter(iv, config_dicts, collect_results=False)
        assert result == []

    def test_run_iter_function_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        iv2 =sel.test_pipeline_manager2.stuff.a_function
        cd = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=10,
            a=2
        )
        cd2 = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=20
        )
        cd3 = dict(
            section_path_str='test_pipeline_manager2.stuff.a_function',
            b=5,
            a=1
        )
        cd4 = dict(
            section_path_str='test_pipeline_manager2.stuff.a_function',
            b=7
        )
        config_dicts = [cd, cd2, cd3, cd4]
        runner = IterativeRunner([iv, iv2], config_dicts)
        result = runner.run()
        assert result == [((cd, cd3), [(2, 10), (1, 5)]), ((cd, cd4), [(2, 10), (None, 7)]),
                          ((cd2, cd3), [(None, 20), (1, 5)]), ((cd2, cd4), [(None, 20), (None, 7)])]
        result = runner.run(collect_results=False)
        assert result == []

    def test_run_iter_function_dependent_on_other_pm_class(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        self.append_to_a_function_config('a = s.test_pipeline_manager2.example_class.stuff.data')
        pipeline_manager.reload()
        sel = Selector()
        expected_a_result = ['a', 'b']
        iv = sel.test_pipeline_manager.stuff.a_function
        iv2 = sel.test_pipeline_manager2.example_class.stuff.data

        # Assert that update pipeline manager 2 affects pipeline manager 1
        cd = dict(
            a=expected_a_result,
            section_path_str=iv2.section_path_str
        )
        config_dicts = [cd]
        runner = IterativeRunner(iv, config_dicts)
        result = runner.run()
        assert result == [((cd,), (iv2, None))]
        assert iv2.a == expected_a_result
        assert context.config_dependencies == self.expect_pm_1_a_function_depends_on_pm_2_specific_class



from copy import deepcopy

from pyfileconf import Selector
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST


class TestPipelineManagerGetOne(PipelineManagerTestBase):

    def test_get_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        iv_func = pipeline_manager.get(iv)
        iv_result = iv_func()
        str_func = pipeline_manager.get('stuff.a_function')
        str_result = str_func()
        assert iv_result == str_result == (None, None)

    def test_get_function_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Get from pipeline manager 1
        iv = sel.test_pipeline_manager.stuff.a_function
        iv_func = pipeline_manager.get(iv)
        iv_result = iv_func()
        str_func = pipeline_manager.get('stuff.a_function')
        str_result = str_func()
        assert iv_result == str_result == (None, None)

        # Get from pipeline manager 2
        iv = sel.test_pipeline_manager2.stuff.a_function
        iv_func = pipeline_manager2.get(iv)
        iv_result = iv_func()
        str_func = pipeline_manager2.get('stuff.a_function')
        str_result = str_func()
        assert iv_result == str_result == (None, None)

    def test_get_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        iv_class = pipeline_manager.get(iv)
        iv_obj = iv_class()
        str_class = pipeline_manager.get('stuff.ExampleClass')
        str_obj = str_class()
        assert iv_obj == str_obj == ExampleClass(None)

    def test_get_class_multiple_pms(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Get from pipeline manager 1
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        iv_class = pipeline_manager.get(iv)
        iv_obj = iv_class()
        str_class = pipeline_manager.get('stuff.ExampleClass')
        str_obj = str_class()
        assert iv_obj == str_obj == ExampleClass(None)

        # Get from pipeline manager 2
        iv = sel.test_pipeline_manager2.stuff.ExampleClass
        iv_class = pipeline_manager2.get(iv)
        iv_obj = iv_class()
        str_class = pipeline_manager2.get('stuff.ExampleClass')
        str_obj = str_class()
        assert iv_obj == str_obj == ExampleClass(None)

    def test_get_class_from_specific_config_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        iv_obj = pipeline_manager.get(iv)
        str_obj = pipeline_manager.get('example_class.stuff.data')
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

    def test_get_class_from_specific_config_dict_multiple_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Get from pipeline manager 1
        iv = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        iv_obj = pipeline_manager.get(iv)
        str_obj = pipeline_manager.get('example_class.stuff.data')
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

        # Get from pipeline manager 2
        iv = sel.test_pipeline_manager2.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        iv_obj = pipeline_manager2.get(iv)
        str_obj = pipeline_manager2.get('example_class.stuff.data')
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

    def test_consistent_specific_config_obj(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        iv_obj = pipeline_manager.get(iv)
        str_obj = pipeline_manager.get('example_class.stuff.data')
        assert iv.item is iv_obj is str_obj

    def test_get_class_from_specific_config_dict_access_property_that_needs_obj_loaded(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        pipeline_manager.update(
            a=(1, 2),
            section_path_str='example_class.stuff.data'
        )
        iv_obj = pipeline_manager.get(iv)
        str_obj = pipeline_manager.get('example_class.stuff.data')
        assert iv.e == iv_obj.e == str_obj.e == 10


class TestPipelineManagerGetSection(PipelineManagerTestBase):

    def test_get_main_dict_section(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff
        iv_section = pipeline_manager.get(iv)
        iv_func = iv_section[0]
        iv_result = iv_func()
        str_section = pipeline_manager.get('stuff')
        str_func = str_section[0]
        str_result = str_func()
        assert iv_result == str_result == (None, None)

    def test_get_main_dict_section_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Get pipeline manager 1 section
        iv = sel.test_pipeline_manager.stuff
        iv_section = pipeline_manager.get(iv)
        iv_func = iv_section[0]
        iv_result = iv_func()
        str_section = pipeline_manager.get('stuff')
        str_func = str_section[0]
        str_result = str_func()
        assert iv_result == str_result == (None, None)

        # Get pipeline manager 2 section
        iv = sel.test_pipeline_manager2.stuff
        iv_section = pipeline_manager2.get(iv)
        iv_func = iv_section[0]
        iv_result = iv_func()
        str_section = pipeline_manager2.get('stuff')
        str_func = str_section[0]
        str_result = str_func()
        assert iv_result == str_result == (None, None)

    def test_get_main_dict_nested_section(self):
        self.write_a_function_to_pipeline_dict_file(nest_section=True)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.my_section
        iv_section = pipeline_manager.get(iv)
        iv_func = iv_section['stuff'][0]
        iv_result = iv_func()
        str_section = pipeline_manager.get('my_section')
        str_func = str_section['stuff'][0]
        str_result = str_func()
        assert iv_result == str_result == (None, None)

    def test_get_specific_class_dict_section(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff
        expect_ec = ExampleClass(None, name='data')
        iv_section = pipeline_manager.get(iv)
        iv_obj = iv_section[0]
        str_section = pipeline_manager.get('example_class.stuff')
        str_obj = str_section[0]
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

    def test_get_specific_class_dict_section_multiple_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Get pipeline manager 1 section
        iv = sel.test_pipeline_manager.example_class.stuff
        expect_ec = ExampleClass(None, name='data')
        iv_section = pipeline_manager.get(iv)
        iv_obj = iv_section[0]
        str_section = pipeline_manager.get('example_class.stuff')
        str_obj = str_section[0]
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

        # Get pipeline manager 2 section
        iv = sel.test_pipeline_manager2.example_class.stuff
        expect_ec = ExampleClass(None, name='data')
        iv_section = pipeline_manager2.get(iv)
        iv_obj = iv_section[0]
        str_section = pipeline_manager2.get('example_class.stuff')
        str_obj = str_section[0]
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

    def test_get_specific_class_dict_nested_section(self):
        self.write_example_class_dict_to_file(nest_section=True)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class
        expect_ec = ExampleClass(None, name='data')
        iv_section = pipeline_manager.get(iv)
        iv_obj = iv_section['my_section']['stuff'][0]
        str_section = pipeline_manager.get('example_class')
        str_obj = str_section['my_section']['stuff'][0]
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a

    def test_get_specific_class_dict_custom_key_attr_section(self):
        self.write_example_class_dict_to_file()
        class_config_dict_list = deepcopy(CLASS_CONFIG_DICT_LIST)
        class_config_dict_list[0].update(
            key_attr='a'
        )
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=class_config_dict_list
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff
        expect_ec = ExampleClass(a='data')
        iv_section = pipeline_manager.get(iv)
        iv_obj = iv_section[0]
        str_section = pipeline_manager.get('example_class.stuff')
        str_obj = str_section[0]
        assert iv_obj.name == str_obj.name == expect_ec.name
        assert iv_obj.a == str_obj.a == expect_ec.a


class TestGetSectionPathFromItem(PipelineManagerTestBase):

    def test_get_function_section_path(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        iv_run = iv()  # result of a_function(), should not have _section_path_str
        assert not hasattr(iv_run, '_section_path_str')
        iv_func = pipeline_manager.get(iv)
        str_func = pipeline_manager.get('stuff.a_function')
        for obj in [iv, iv_func, str_func]:
            sp = obj._section_path_str
            assert sp == 'test_pipeline_manager.stuff.a_function'

    def test_get_function_section_path_by_section(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff
        iv_func = pipeline_manager.get(iv)
        str_func = pipeline_manager.get('stuff')
        assert iv._section_path_str == 'test_pipeline_manager.stuff'
        for obj_list in [iv_func, str_func]:
            for obj in obj_list:
                sp = obj._section_path_str
                assert sp == 'test_pipeline_manager.stuff.a_function'

    def test_get_class_section_path(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        iv_run = iv()  # running general class gets instance of class, so should have _section_path_str
        iv_class = pipeline_manager.get(iv)
        str_class = pipeline_manager.get('stuff.ExampleClass')
        for obj in [iv, iv_run, iv_class, str_class]:
            sp = obj._section_path_str
            assert sp == 'test_pipeline_manager.stuff.ExampleClass'

    def test_get_class_section_path_by_section(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff
        iv_class = pipeline_manager.get(iv)
        str_class = pipeline_manager.get('stuff')
        assert iv._section_path_str == 'test_pipeline_manager.stuff'
        for obj_list in [iv_class, str_class]:
            for obj in obj_list:
                sp = obj._section_path_str
                assert sp == 'test_pipeline_manager.stuff.ExampleClass'

    def test_get_specific_class_section_path(self):
        self.write_example_class_dict_to_file()
        ccdl = deepcopy(CLASS_CONFIG_DICT_LIST)
        for config_dict in ccdl:
            config_dict['execute_attr'] = 'return_section_path_str'
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=ccdl
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        # result of __call__ on ExampleClass, should not have _section_path_str on result, but should have in object
        iv_run = iv()
        pm_run = pipeline_manager.run(iv)
        assert not hasattr(iv_run, '_section_path_str')
        assert iv_run == pm_run == 'test_pipeline_manager.example_class.stuff.data'
        iv_obj = pipeline_manager.get(iv)
        str_obj = pipeline_manager.get('example_class.stuff.data')
        for obj in [iv, iv_obj, str_obj]:
            sp = obj._section_path_str
            assert sp == 'test_pipeline_manager.example_class.stuff.data'

    def test_get_specific_class_section_path_by_section(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff
        iv_obj = pipeline_manager.get(iv)
        str_obj = pipeline_manager.get('example_class.stuff')
        assert iv._section_path_str == 'test_pipeline_manager.example_class.stuff'
        for obj_list in [iv_obj, str_obj]:
            for obj in obj_list:
                sp = obj._section_path_str
                assert sp == 'test_pipeline_manager.example_class.stuff.data'


from ast import AST
import os
from typing import List

from mixins.propertycache import SimplePropertyCacheMixin
from pyfileconf.io.file.load.parsers.py import PythonFileParser
from mixins.repr import ReprMixin

class LazyLoader(SimplePropertyCacheMixin, ReprMixin):
    repr_cols = ['filepath']

    def __init__(self, filepath: str):
        self.filepath = filepath

    def load(self):
        pass

    def register(self):
        if os.path.exists(self.filepath):
            self._ast, self._body = PythonFileParser(self.filepath).load()
        else:
            self._ast, self._body = None, None

    @property
    def ast(self) -> AST:
        return self._try_getattr_else_register('_ast')

    @property
    def body(self) -> List[str]:
        return self._try_getattr_else_register('_body')

    def _try_getattr_else_register(self, attr: str):
        return self._try_getattr_else_call_func(attr, self.register)


import ast
from typing import List, Union

FunctionDefOrNone = Union[ast.FunctionDef, None]
FunctionDefs = List[ast.FunctionDef]

class FunctionDefinitionExtractor(ast.NodeVisitor):

    def __init__(self):
        self.defs = []

    def visit_FunctionDef(self, node):
        self.defs.append(node)
        # don't go into children, so won't extract nested functions

    def visit_ClassDef(self, node):
        pass
        # don't go into children, so won't extract class methods


def extract_function_definitions_from_ast(module: ast.AST) -> FunctionDefs:
    fe = FunctionDefinitionExtractor()
    fe.visit(module)
    return fe.defs

class FunctionDefinitionByNameExtractor(FunctionDefinitionExtractor):

    def __init__(self, func_name: str):
        self.func_name = func_name
        super().__init__()

    def visit_FunctionDef(self, node):
        if node.name == self.func_name:
            super().visit_FunctionDef(node)

def extract_function_definitions_from_ast_by_name(module: ast.AST, name: str) -> FunctionDefs:
    fe = FunctionDefinitionByNameExtractor(name)
    fe.visit(module)
    return fe.defs

def extract_function_definition_from_ast_by_name(module: ast.AST, name: str) -> FunctionDefOrNone:
    defs = extract_function_definitions_from_ast_by_name(module, name)

    if len(defs) == 0:
        return None

    if len(defs) > 1:
        raise ValueError(f'found more than one function definition with the name {name} in module {module}')

    return defs[0]

class FunctionDefinitionOrClassInitByNameExtractor(FunctionDefinitionByNameExtractor):

    def visit_ClassDef(self, node):
        if node.name == self.func_name:
            # Found a class definition matching the name we're looking for
            # Now look for the init method of that class
            # TODO [#11]: deal with subclassing where __init__ will not be in class definition
            orig_name = self.func_name
            self.func_name = '__init__'
            self.generic_visit(node)
            self.func_name = orig_name
            # Set back to looking for original name. Don't want to pull other inits

def extract_function_definitions_or_class_inits_from_ast_by_name(module: ast.AST, name: str) -> FunctionDefs:
    fe = FunctionDefinitionOrClassInitByNameExtractor(name)
    fe.visit(module)
    return fe.defs

def extract_function_definition_or_class_init_from_ast_by_name(module: ast.AST, name: str) -> FunctionDefOrNone:
    defs = extract_function_definitions_or_class_inits_from_ast_by_name(module, name)

    if len(defs) == 0:
        return None

    if len(defs) > 1:
        raise ValueError(f'found more than one function definition with the name {name} in module {module}')

    return defs[0]



from pyfileconf.sectionpath.sectionpath import SectionPath


def _set_in_nested_obj_by_section_path(obj, section_path: SectionPath, value) -> None:
    # Goes into nested sections, until it sets the final section or pipeline
    for section in section_path[:-1]:
        obj = getattr(obj, section)

    # Now have object of next level up from object desired to be set
    setattr(obj, section_path[-1], value)



from typing import Union
import ast

from pyfileconf.imports.models.statements.module import ModuleImportStatement
from pyfileconf.imports.models.statements.obj import ObjectImportStatement
from pyfileconf.imports.models.statements.comment import Comment

AnyAstImport = Union[ast.Import, ast.ImportFrom]
AnyImportStatement = Union[ModuleImportStatement, ObjectImportStatement]
AnyImportStatementOrComment = Union[AnyImportStatement, Comment]
ImportOrNone = Union[AnyImportStatement, None]


import ast

class KeywordExtractor(ast.NodeVisitor):

    def __init__(self):
        self.kwargs = {}

    def visit_keyword(self, node):
        self.kwargs[node.arg] = node.value
        self.generic_visit(node)


def extract_keywords_from_ast(node: ast.AST) -> dict:
    ke = KeywordExtractor()
    ke.visit(node)
    return ke.kwargs


class KeywordByNameExtractor(KeywordExtractor):

    def __init__(self, name: str):
        self.name = name
        super().__init__()

    def visit_keyword(self, node):
        if node.arg == self.name:
            super().visit_keyword(node)


def extract_keywords_from_ast_by_name(node: ast.AST, name: str) -> dict:
    ke = KeywordByNameExtractor(name)
    ke.visit(node)
    return ke.kwargs


import sys

class ImportTracker:

    def __init__(self):
        self.original_modules = list(sys.modules.keys())

    @property
    def imported_modules(self):
        current_modules = list(sys.modules.keys())
        return [module for module in current_modules if module not in self.original_modules]





from pyfileconf.io.file.load.active.loader import ActiveConfigFileLoader

class ActiveConfigFileInterface(ActiveConfigFileLoader):

    def save(self):
        raise NotImplementedError('call save with ConfigFileInterface, not ActiveConfigFileInterface')



from pyfileconf.pipelines.models.config import FunctionConfig
from pyfileconf.config.models.file import ActiveFunctionConfigFile

class ActiveFunctionConfig(FunctionConfig):

    config_file_class = ActiveFunctionConfigFile


import os
from copy import deepcopy

from pyfileconf import Selector, PipelineManager
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.context import context
from tests.input_files.amodule import SecondExampleClass
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST, SAME_CLASS_CONFIG_DICT_LIST, \
    DIFFERENT_CLASS_CONFIG_DICT_LIST


class TestPipelineManagerConfig(PipelineManagerTestBase):

    def test_config_update_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        expected_b_result = ['a', 'b']
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager.run(iv)
        assert result == (None, expected_b_result)

    def test_config_update_by_file_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        expected_b_result = ['a', 'b']
        b_str = f'b = {expected_b_result}'
        self.append_to_a_function_config(b_str)
        pipeline_manager.reload()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, expected_b_result)

    def test_config_update_function_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        expected_b_result = ['a', 'b']
        iv = sel.test_pipeline_manager.stuff.a_function
        iv2 = sel.test_pipeline_manager2.stuff.a_function

        # Assert update pipeline manager 1
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager.run(iv)
        assert result == (None, expected_b_result)

        # Assert that pipeline manager 2 is not updated yet
        result2 = pipeline_manager2.run(iv2)
        assert result2 == (None, None)

        # Assert update pipeline manager 2
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager2.run(iv2)
        assert result == (None, expected_b_result)

    def test_config_update_function_multiple_dependent_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        self.append_to_a_function_config('b = s.test_pipeline_manager2.stuff.a_function')
        pipeline_manager.reload()
        sel = Selector()
        expected_b_result = ['a', 'b']
        iv = sel.test_pipeline_manager.stuff.a_function
        iv2 = sel.test_pipeline_manager2.stuff.a_function

        # Assert original pipeline manager 1 has pipeline manager 2 a_function as b
        result = pipeline_manager.run(iv)
        assert result == (None, iv2)

        # Assert that update pipeline manager 2 affects pipeline manager 1
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager.run(iv2)
        assert result == (None, iv2)
        assert result[1]() == (None, expected_b_result)
        assert context.config_dependencies == self.expect_pm_1_a_function_depends_on_pm_2_a_function

    def test_config_update_function_dependencies(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        pipeline_manager.update(
            section_path_str='stuff.a_function',
            b=sel.test_pipeline_manager2.stuff.a_function
        )
        assert context.config_dependencies == self.expect_pm_1_a_function_depends_on_pm_2_a_function

    def test_config_update_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(expected_a_result)

    def test_config_update_by_file_for_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        expected_a_result = (1, 2)
        a_str = f'a = {expected_a_result}'
        self.append_to_example_class_config(a_str)
        pipeline_manager.reload()
        sel = Selector()
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(expected_a_result)

    def test_create_update_class_multiple_pms(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        expected_a_result = (1, 2)
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        iv2 = sel.test_pipeline_manager2.stuff.ExampleClass

        # Assert update pipeline manager 1
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(expected_a_result)

        # Assert that pipeline manager 2 is not updated yet
        ec = sel.test_pipeline_manager2.stuff.ExampleClass()
        assert ec == ExampleClass(None)

        # Assert update pipeline manager 2
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec = sel.test_pipeline_manager2.stuff.ExampleClass()
        assert ec == ExampleClass(expected_a_result)

    def test_create_update_class_multiple_dependent_pms(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        self.append_to_example_class_config('a = s.test_pipeline_manager2.stuff.ExampleClass')
        pipeline_manager.reload()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        iv2 = sel.test_pipeline_manager2.stuff.ExampleClass

        # Assert original pipeline manager 1 has pipeline manager 2 example class as a
        ec2 = sel.test_pipeline_manager2.stuff.ExampleClass
        expect_1 = ExampleClass(ec2)
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == expect_1
        assert ec.a().a is None

        # Assert that update pipeline manager 2 affects pipeline manager 1
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec2 = sel.test_pipeline_manager2.stuff.ExampleClass
        expect_1 = ExampleClass(ec2)
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == expect_1
        assert ec.a().a == expected_a_result
        assert context.config_dependencies == self.expect_pm_1_class_depends_on_pm_2_class

    def test_config_update_class_dependencies(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        pipeline_manager.update(
            section_path_str='stuff.ExampleClass',
            a=sel.test_pipeline_manager2.stuff.ExampleClass
        )
        assert context.config_dependencies == self.expect_pm_1_class_depends_on_pm_2_class


    def test_create_update_from_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_config_update_by_file_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        expected_a_result = (1, 2)
        a_str = f'a = {expected_a_result}'
        self.append_to_specific_class_config(a_str)
        pipeline_manager.reload()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_create_update_from_specific_class_dict_multiple_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()
        expected_a_result = (1, 2)
        iv = sel.test_pipeline_manager.example_class.stuff.data
        iv2 = sel.test_pipeline_manager2.example_class.stuff.data

        # Assert update pipeline manager 1
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

        # Assert that pipeline manager 2 is not updated yet
        ec = sel.test_pipeline_manager2.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

        # Assert update pipeline manager 2
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec = sel.test_pipeline_manager2.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_create_update_from_specific_class_dict_multiple_dependent_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        self.append_to_specific_class_config('a = s.test_pipeline_manager2.example_class.stuff.data')
        sel = Selector()
        iv2 = sel.test_pipeline_manager2.example_class.stuff.data

        # Assert original pipeline manager 1 has pipeline manager 2 example class as a
        ec2 = sel.test_pipeline_manager2.example_class.stuff.data
        expect_1 = ExampleClass(name='data', a=ec2)
        ec = sel.test_pipeline_manager.example_class.stuff.data.item
        assert ec == expect_1
        assert ec.a.a is None

        # Assert that update pipeline manager 2 affects pipeline manager 1
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec2 = sel.test_pipeline_manager2.example_class.stuff.data
        expect_1 = ExampleClass(name='data', a=ec2)
        ec = sel.test_pipeline_manager.example_class.stuff.data.item
        assert ec == expect_1
        assert ec.a.a == expected_a_result
        assert context.config_dependencies == self.expect_pm_1_specific_class_depends_on_pm_2_specific_class

    def test_create_update_specific_class_dependencies(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()
        pipeline_manager.update(
            section_path_str='example_class.stuff.data',
            a=sel.test_pipeline_manager2.example_class.stuff.data,
        )
        assert context.config_dependencies == self.expect_pm_1_specific_class_depends_on_pm_2_specific_class

    def test_create_update_from_specific_class_dict_multiple_dependent_attribute_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        self.append_to_specific_class_config('a = s.test_pipeline_manager2.example_class.stuff.data.a')
        pipeline_manager.reload()
        sel = Selector()
        iv2 = sel.test_pipeline_manager2.example_class.stuff.data

        # Assert original pipeline manager 1 has pipeline manager 2 example class as a
        ec2 = sel.test_pipeline_manager2.example_class.stuff.data
        expect_1 = ExampleClass(name='data', a=ec2.a)
        ec = sel.test_pipeline_manager.example_class.stuff.data.item
        assert ec == expect_1
        assert ec.a is None

        # Assert that update pipeline manager 2 affects pipeline manager 1
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        ec2 = sel.test_pipeline_manager2.example_class.stuff.data
        expect_1 = ExampleClass(name='data', a=ec2.a)
        ec = sel.test_pipeline_manager.example_class.stuff.data.item
        assert ec == expect_1
        assert ec.a == expected_a_result
        assert context.config_dependencies == context.force_update_dependencies == \
               self.expect_pm_1_specific_class_depends_on_pm_2_specific_class

    def test_create_update_from_multiple_specific_class_dicts_same(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(1)  # example_class2
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=SAME_CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        ec2 = sel.test_pipeline_manager.example_class2.stuff.data
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(ec.section_path_str)[1:])
        section_path2 = SectionPath.from_section_str_list(SectionPath(ec2.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path2.path_str
        )
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == ec2.name == expect_ec.name
        assert ec.a == ec2.a == expect_ec.a

    def test_create_update_from_multiple_specific_class_dicts_different(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(2)  # second_example_class
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=DIFFERENT_CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        sec = sel.test_pipeline_manager.second_example_class.stuff.data
        expected_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(ec.section_path_str)[1:])
        section_path_s = SectionPath.from_section_str_list(SectionPath(sec.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.update(
            b=expected_result,
            section_path_str=section_path_s.path_str
        )
        expect_ec = ExampleClass(name='data', a=expected_result)
        expect_sec = SecondExampleClass(name='data', b=expected_result)
        assert ec.name == sec.name == expect_ec.name
        assert ec.a == sec.b == expect_ec.a == expect_sec.b

    def test_config_reload_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        expected_b_result = ['a', 'b']
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.reload()
        result = pipeline_manager.run(iv)
        assert result == (None, None)

    def test_config_reload_function_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        iv2 = sel.test_pipeline_manager2.stuff.a_function
        expected_b_result = ['a', 'b']

        # Update both pipeline manager configs
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )

        # Assert that reloading pipeline manager 1 resets its config
        pipeline_manager.reload()
        result = pipeline_manager.run(iv)
        assert result == (None, None)

        # Assert that the reload of pipeline manager 1 did not affect pipeline manager 2
        result = pipeline_manager2.run(iv2)
        assert result == (None, expected_b_result)

    def test_config_reload_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.reload()
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(None)

    def test_config_reload_class_multiple_pms(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()
        expected_a_result = (1, 2)
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        iv2 = sel.test_pipeline_manager2.stuff.ExampleClass

        # Update both pipeline manager configs
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )

        # Assert that reloading pipeline manager 1 resets its config
        pipeline_manager.reload()
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(None)

        # Assert that the reload of pipeline manager 1 did not affect pipeline manager 2
        ec = sel.test_pipeline_manager2.stuff.ExampleClass()
        assert ec == ExampleClass(expected_a_result)

    def test_config_reload_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.reload()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_config_reload_specific_class_dict_multiple_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()
        expected_a_result = (1, 2)
        iv = sel.test_pipeline_manager.example_class.stuff.data
        iv2 = sel.test_pipeline_manager2.example_class.stuff.data

        # Update both pipeline manager configs
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        section_path = SectionPath.from_section_str_list(SectionPath(iv2.section_path_str)[1:])
        pipeline_manager2.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )

        # Assert that reloading pipeline manager 1 resets its config
        pipeline_manager.reload()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

        # Assert that the reload of pipeline manager 1 did not affect pipeline manager 2
        ec = sel.test_pipeline_manager2.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_config_update_class_used_by_function_through_selector(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager.load()

        # Link function to specific class object
        b_str = 'b = s.test_pipeline_manager.example_class.stuff.data'
        self.append_to_a_function_config(b_str)

        # Update specific class object config file to ensure we are getting the correct one later
        expected_a_result = (1, 2)
        a_str = f'a = {expected_a_result}'
        self.append_to_specific_class_config(a_str)

        pipeline_manager.reload()
        sel = Selector()

        # Assert that we can already get the specific class object returned before updating config
        f_iv = sel.test_pipeline_manager.stuff.a_function
        sc_iv = sel.test_pipeline_manager.example_class.stuff.data
        result = pipeline_manager.run(f_iv)
        sc = pipeline_manager.get(sc_iv)
        expect_sc = ExampleClass(a=expected_a_result, name='data')
        result_none, result_sc = result
        assert result_none is None
        assert result_sc.name == sc.name == expect_sc.name
        assert result_sc.a == sc.a == expect_sc.a

        # Assert that updating the specific class object updates the result from the function
        second_expected_a_result = (3, 4)
        section_path = SectionPath.from_section_str_list(SectionPath(sc_iv.section_path_str)[1:])
        pipeline_manager.update(
            a=second_expected_a_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager.run(f_iv)
        sc = pipeline_manager.get(sc_iv)
        expect_sc = ExampleClass(a=second_expected_a_result, name='data')
        result_none, result_sc = result
        assert result_none is None
        assert result_sc.name == sc.name == expect_sc.name
        assert result_sc.a == sc.a == expect_sc.a
        assert context.config_dependencies == self.expect_pm_1_a_function_depends_on_pm_1_specific_class

    def test_config_update_function_by_item_view(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        expected_b_result = ['a', 'b']
        iv.b = expected_b_result
        result = pipeline_manager.run(iv)
        assert result == (None, expected_b_result)

    def test_config_update_class_by_item_view(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        expected_a_result = (1, 2)
        iv.a = expected_a_result
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(expected_a_result)

    def test_update_specific_class_config_attr_by_item_view(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        expected_a_result = (1, 2)
        iv.a = expected_a_result
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_deep_copy_specific_class_dict_item_view(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        new_item = deepcopy(iv)
        assert isinstance(new_item, ExampleClass)
        assert iv.item is not new_item

    def test_invalid_config_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        self.write_error_to_a_function_file()

        # No error raised here thanks to lazy loading
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function

        # Config is only loaded once attribute of item view is accessed
        with self.assertRaises(ValueError):
            iv.item

    def test_invalid_config_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        self.write_error_to_example_class_file()

        # No error raised here thanks to lazy loading
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass

        # Config is only loaded once attribute of item view is accessed
        with self.assertRaises(ValueError):
            iv.item

    def test_invalid_specific_config_class(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        self.write_error_to_specific_example_class_file()

        # No error raised here thanks to lazy loading
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data

        # Config is only loaded once attribute of item view is accessed
        with self.assertRaises(ValueError):
            iv.item


from typing import List

class FileParser:

    def __init__(self, filepath: str):
        self.filepath = filepath

    def load(self) -> List[str]:
        with open(self.filepath, 'r', encoding='utf8') as f:
            contents = f.readlines()

        return contents


from typing import TYPE_CHECKING, Sequence, cast

from pyfileconf.exceptions.imports import CouldNotExtractRenameException

if TYPE_CHECKING:
    from pyfileconf.imports.models.statements.interfaces import AnyAstImport
import ast

from mixins.repr import ReprMixin
from mixins.attrequals import EqOnAttrsMixin

class RenameStatement(ReprMixin, EqOnAttrsMixin):
    repr_cols = ['item', 'new_name']
    equal_attrs = ['item', 'new_name']

    def __init__(self, item: str, new_name: str):
        self.item = item
        self.new_name = new_name

    def __str__(self):
        return f'{self.item} as {self.new_name}'

    @classmethod
    def from_ast_alias(cls, alias: ast.alias):
        # Note: will fail if ast alias does not have a rename. See RenameStatementCollection.from_ast_import
        if alias.asname is None:
            raise CouldNotExtractRenameException(f'must pass alias which as .asname. passed alias: {alias}')

        return cls(
            alias.name,
            alias.asname
        )


class RenameStatementCollection(ReprMixin):
    repr_cols = ['items']

    def __init__(self, items: Sequence[RenameStatement]):
        self.items = items

    def __contains__(self, item):
        if isinstance(item, RenameStatement):
            return item in self.items

        # If not rename statement, assuming got variable name
        return item in [rename_statement.item for rename_statement in self.items]

    def __getitem__(self, item):
        return self.item_map[item]

    def __iter__(self):
        for item in self.items:
            yield item

    @property
    def new_names(self):
        return [item.new_name for item in self]

    @property
    def orig_names(self):
        return [item.item for item in self]

    @property
    def item_map(self):
        return {rename_statement.item: rename_statement for rename_statement in self.items}

    @property
    def reverse_name_map(self):
        return {rename_statement.new_name: rename_statement.item for rename_statement in self.items}

    @property
    def name_map(self):
        return {rename_statement.item: rename_statement.new_name for rename_statement in self.items}

    @classmethod
    def from_ast_import(cls, ast_import: 'AnyAstImport'):
        # For ast aliases, they always exist whether the item is being renamed or not.
        # For RenameStatement objects, they only exist when there is a rename. Need to filter
        renames = []
        for alias in ast_import.names:
            alias = cast(ast.alias, alias)
            if alias.asname is not None:
                renames.append(RenameStatement.from_ast_alias(alias))

        return cls(renames)


from copy import deepcopy
from typing import Union, cast, Any, Type
import os

from pyfileconf.basemodels.collection import Collection
from pyfileconf.data.logic.convert import convert_to_empty_obj_if_necessary
from pyfileconf.logic.get import _get_public_name_or_special_name
from pyfileconf.data.models.config import SpecificClassConfig
from pyfileconf.io.func.load.args import extract_function_args_and_arg_imports_from_import
from pyfileconf.imports.models.statements.obj import ObjectImportStatement
from pyfileconf.imports.logic.load.name import get_module_and_name_imported_from
from pyfileconf.data.models.astitems import ast_str
from pyfileconf.io.func.load.config import function_args_as_arg_and_annotation_dict


ObjOrCollection = Union[Any, 'SpecificClassCollection']

class SpecificClassCollection(Collection):
    klass: Type

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.klass is None:
            raise ValueError('must pass class for SpecificClassCollection')

    def _set_name_map(self) -> None:
        obj_map = {}
        for obj_or_collection in self:
            obj_or_collection = cast(ObjOrCollection, obj_or_collection)
            if isinstance(obj_or_collection, SpecificClassCollection):
                key_attr = 'name'
            else:
                key_attr = self.key_attr
            obj_name = getattr(obj_or_collection, key_attr)
            obj_map[obj_name] = obj_or_collection
        self.name_dict = obj_map

    def _transform_item(self, item):
        """
        Is called on each item when adding items to collection. Should handle whether the item
        is an actual item or another collection. Must return the item or collection.

        If not overridden, will just store items as is.

        Returns: item or Collection

        """
        return convert_to_empty_obj_if_necessary(item, self.klass, key_attr=self.key_attr)

    def _output_config_files(self) -> None:
        if not os.path.exists(self.basepath):
            os.makedirs(self.basepath)

        for item in self:
            self._output_config_file(item)

    def _output_config_file(self, item: ObjOrCollection) -> None:
        if isinstance(item, SpecificClassCollection):
            # if collection, recursively call creating config files
            return item._output_config_files()

        # Dealing with object itself
        item_name = getattr(item, self.key_attr)
        item_filepath = os.path.join(self.basepath, item_name + '.py')

        class_config = dict(
            klass=self.klass,
            always_import_strs=self.always_import_strs,
            always_assign_strs=self.always_assign_strs
        )

        if os.path.exists(item_filepath):
            file_existed = True
            # if config file already exists, load confguration from file, use to update file defaults
            existing_config = SpecificClassConfig.from_file(
                item_filepath,
                name=item_name,
                **class_config,  # type: ignore
            )
            existing_imports = existing_config._file.interface.imports
            # Here using the ast config, for the purpose of writing to file
            file_config_item = deepcopy(item)
            apply_config(file_config_item, existing_config)
        else:
            file_existed = False
            existing_imports = None
            file_config_item = item

        item_config = SpecificClassConfig(imports=existing_imports,
                                                   file_path=item_filepath,
                                                   **class_config)  # type: ignore

        # Get config by extracting from class __init__
        # First need to create dummy import for compatibility
        mod, import_base = get_module_and_name_imported_from(self.klass)
        obj_import = ObjectImportStatement([item_name], import_base)
        # Now get the arguments and the imports for any type annotations
        args, func_arg_imports = extract_function_args_and_arg_imports_from_import(
            self.klass.__name__,
            obj_import
        )
        # Convert into a usable formats:
        # defaults_dict: a dictionary where keys are variable names and values are ast defaults
        # annotation_dict: a dicrionary where keys are variable names and values are ast type annotations
        defaults_dict, annotation_dict = function_args_as_arg_and_annotation_dict(args)
        defaults_dict[self.key_attr] = ast_str(item_name)  # set name attribute as item name by default
        # Apply all the new extracted defaults to the created config
        item_config.update(defaults_dict)
        item_config.annotations.update(annotation_dict)
        item_config.imports.extend(func_arg_imports)

        item_config.to_file(item_filepath)


def apply_config(obj: Any, config: 'SpecificClassConfig') -> None:
    attributes = dir(obj)
    for config_attr, config_item in config.items():
        # Skip irrelevant items
        if config_attr in attributes:
            setattr(obj, config_attr, config_item)


from typing import List
from pyfileconf.sectionpath.sectionpath import SectionPath

accepted_name_attrs = ['name', '__name__']
output_accepted_name_attrs = ['output_name'] + accepted_name_attrs


def _get_from_nested_obj_by_section_path(obj, section_path: SectionPath, prevent_property_access: bool = False):
    """

    :param obj:
    :param section_path:
    :param prevent_property_access: Whether to prevent accessing properties. This is used in selector as the
        selector is working with the unconfigured objects for lookup, and the unconfigured objects' properties
        could cause errors or other side effects
    :return:
    """
    # Goes into nested sections, until it pulls the final section or pipeline
    for section in section_path:

        # When preventing property access, checks if next accessed attribute is actually a property
        try:
            if prevent_property_access and isinstance(getattr(type(obj), section), property):
                # Next section would access a property, which could lead to unintended side-effects
                # Therefore, return an object which is not a collection, function, or specific class object,
                # as this will signal to calling functions that the full section path is not a stored object

                # the returned value is totally arbitrary so long as it is not a collection,
                # function, or specific class object
                return 5
        except AttributeError:
            # If there is no property to look up under this name, then just continue as normal.
            # Not exactly sure why this was needed since the next code will raise the same error, but
            # for some reason it won't work without this try/except block
            pass

        obj = getattr(obj, section)

    return obj


def _get_public_name_or_special_name(obj, accept_output_names=True) -> str:
    name_attrs = _get_name_attrs(accept_output=accept_output_names)
    for name_var in name_attrs:
        if hasattr(obj, name_var):
            return getattr(obj, name_var)

    raise ValueError(f'could not get .name or .__name__ from {obj} of type {type(obj)}')

def _has_public_name_or_special_name(obj, accept_output_names=True) -> bool:
    name_attrs = _get_name_attrs(accept_output=accept_output_names)
    for name_var in name_attrs:
        if hasattr(obj, name_var):
            return True

    return False

def _get_name_attrs(accept_output: bool=True) -> List[str]:
    if accept_output:
        return output_accepted_name_attrs
    else:
        return accepted_name_attrs


from typing import Tuple, Optional, Type, Sequence, Any, Dict
import os
from copy import deepcopy

from pyfileconf.basemodels.file import ConfigFileBase
from pyfileconf.imports.models.statements.container import ImportStatementContainer
from pyfileconf.assignments.models.container import AssignmentStatementContainer

ImportsAndAssigns = Tuple[ImportStatementContainer, AssignmentStatementContainer]

class ConfigBase(dict):

    ##### Scaffolding functions or attributes. Need to override when subclassing  ####

    config_file_class = ConfigFileBase

    ##### Base class functions and attributes below. Shouldn't usually need to override in subclassing #####

    def __init__(self, d: dict=None, name: str=None, annotations: dict=None, imports: ImportStatementContainer = None,
                 _file: ConfigFileBase=None, begin_assignments: AssignmentStatementContainer=None,
                 klass: Optional[Type] = None, always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None,
                 **kwargs):
        if d is None:
            d = {}
        super().__init__(d, **kwargs)

        if annotations is None:
            annotations = {}

        if imports is None:
            imports = ImportStatementContainer([])

        if begin_assignments is None:
            begin_assignments = AssignmentStatementContainer([])

        self.name = name
        self.annotations = annotations
        self.imports = imports
        self._file = _file
        self.begin_assignments = begin_assignments
        self.klass = klass
        self.always_import_strs = always_import_strs
        self.always_assign_strs = always_assign_strs
        self._applied_updates: Dict[str, Any] = {}

    def __repr__(self):
        dict_repr = super().__repr__()
        class_name = self.__class__.__name__
        return f'<{class_name}(name={self.name}, {dict_repr})>'

    def __getattr__(self, attr):
        try:
            self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __dir__(self):
        return self.keys()

    # argument names to match dict.update
    def update(self, E=None, **F):
        if E is None:
            E = {}

        # Track the updates so they can be applied later
        all_updates = {**E, **F}
        self._applied_updates.update(all_updates)

        super().update(E, **F)

    def to_file(self, filepath: str):

        if self._file is None:
            output_file = self.config_file_class(
                filepath,
                name=self.name,
                klass=self.klass,
                always_import_strs=self.always_import_strs,
                always_assign_strs=self.always_assign_strs
            )
        else:
            # In case this is a new filepath for the same config, copy old file contents for use in new filepath
            output_file = deepcopy(self._file)
            output_file.filepath = filepath

        if os.path.exists(filepath):
            output_file.load() # load any existing config saved in the file, for preserving of user-saved inputs

        output_file.save(self)

    @classmethod
    def from_file(cls, filepath: str, name: str = None,
                  klass: Optional[Type] = None, always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None):
        file = cls.config_file_class(
            filepath,
            name=name,
            klass=klass,
            always_import_strs=always_import_strs,
            always_assign_strs=always_assign_strs
        )
        return file.load(cls)

    def as_imports_and_assignments(self) -> ImportsAndAssigns:
        assigns = AssignmentStatementContainer.from_dict_of_varnames_and_ast(self, self.annotations)

        return self.imports, self.begin_assignments + assigns

    def copy(self):
        return deepcopy(self)

    def refresh(self):
        """
        Reloads from the existing, then reapplies any config updates. Useful for when
        this config depends on the attribute of some other config which was updated.
        :return:
        """
        # Reload from file
        new_config = self.__class__.from_file(self._file.filepath, name=self.name,
                                              klass=self.klass, always_import_strs=self.always_import_strs,
                                              always_assign_strs=self.always_assign_strs)
        applied_updates = self._applied_updates
        self.update(**new_config)
        self.update(**applied_updates)


from typing import List, Dict
import ast
import importlib
from types import ModuleType

from mixins.repr import ReprMixin

from pyfileconf.exceptions.imports import CouldNotExtractModuleImportFromAstException
from pyfileconf.imports.models.statements.importbase import ImportStatement
from pyfileconf.imports.models.statements.rename import RenameStatementCollection
from pyfileconf.imports.models.statements.comment import Comment
from pyfileconf.imports.logic.load.ext_importlib import get_filepath_from_module_str


class ModuleImportStatement(ImportStatement, ReprMixin):
    rename_attr = 'modules'
    repr_cols = ['modules', 'renames', 'comment']
    equal_attrs = ['modules', 'renames', 'comment']

    def __init__(self, modules: List[str], renames: RenameStatementCollection = None, comment: Comment=None,
                 preferred_position: str = None):

        if renames is None:
            renames = RenameStatementCollection([])

        self.modules = modules
        self.renames = renames
        self.comment = comment
        self.preferred_position = preferred_position  # sets self.prefer_beginning as bool

    def __str__(self):
        modules = self._renamed
        modules_str = ', '.join(modules)
        import_str = f'import {modules_str}'
        if self.comment is not None:
            import_str += f' {self.comment}'

        return import_str

    # TODO [#24]: handle renames in ModuleImportStatement.from_str
    @classmethod
    def from_str(cls, import_str: str, renames: RenameStatementCollection = None, comment: Comment=None,
                 preferred_position: str = None):
        from pyfileconf.io.file.load.parsers.imp import extract_module_import_from_ast
        ast_module = ast.parse(import_str)
        cls_obj = extract_module_import_from_ast(ast_module)
        if cls_obj is None:
            raise CouldNotExtractModuleImportFromAstException(f'Original str which was converted to ast: {import_str}')
        cls_obj.comment = comment
        cls_obj.preferred_position = preferred_position

        return cls_obj

    @classmethod
    def from_ast_import(cls, ast_import: ast.Import, preferred_position: str = None):

        # Create RenameStatementCollection
        renames = RenameStatementCollection.from_ast_import(ast_import)

        # Get module original names
        modules = [alias.name for alias in ast_import.names]

        return cls(
            modules,
            renames,
            preferred_position=preferred_position
        )

    def get_module_filepaths(self, import_section_path_str: str=None) -> Dict[str, str]:
        return {module: get_filepath_from_module_str(module, import_section_path_str) for module in self.modules}

    def execute(self, import_section_path_str: str=None) -> List[ModuleType]:
        return [importlib.import_module(mod_str, import_section_path_str) for mod_str in self.modules]

    @property
    def module(self):
        if len(self.modules) > 1:
            raise ValueError('cannot get one module, import has multiple modules')

        return self.modules[0]



from typing import List, Any, Tuple, Iterator
from types import ModuleType
import sys

from pyfileconf.exceptions.imports import CouldNotDetermineModuleForObjectException
from pyfileconf.imports.logic.load.skipmodules import skip_modules
from pyfileconf.sectionpath.sectionpath import SectionPath


def get_imported_obj_variable_name(obj, module: ModuleType) -> str:
    key_list, value_list = _get_module_keys_and_values_lists(module)
    return _get_key_matching_value(obj, key_list, value_list)


def get_module_and_name_imported_from(obj, search_list: List[str]=None) -> Tuple[ModuleType, str]:
    if search_list is None:
        search_list = list(sys.modules.keys())

    for module_name in search_list:

        # skip modules which were causing issues
        if _should_skip_module(module_name):
            continue

        module = sys.modules[module_name]
        if _obj_in_module(obj, module):
            return module, module_name

    raise CouldNotDetermineModuleForObjectException(f'could not find {obj} in {search_list}')

def is_imported_name(name: str, search_list: List[str]=None) -> bool:
    if search_list is None:
        search_list = list(sys.modules.keys())

    for module_name in search_list:

        # skip modules which were causing issues
        if _should_skip_module(module_name):
            continue

        module = sys.modules[module_name]
        if _name_in_module(name, module):
            return True

    return False

def is_imported_obj(obj, search_list: List[str]=None) -> bool:
    if search_list is None:
        search_list = list(sys.modules.keys())

    for module_name in search_list:

        # skip modules which were causing issues
        if _should_skip_module(module_name):
            continue

        module = sys.modules[module_name]
        if _obj_in_module(obj, module):
            return True

    return False

def _is_imported_from(name: str, search_list: List[str]=None) -> List[str]:
    if search_list is None:
        search_list = list(sys.modules.keys())

    matched_modules = []
    for module_name in search_list:

        # skip modules which were causing issues
        if _should_skip_module(module_name):
            continue

        module = sys.modules[module_name]
        if _name_in_module(name, module):
            matched_modules.append(module_name)

    return matched_modules

def _get_key_matching_value(value, key_list, value_list) -> str:
    for i, match_value in enumerate(value_list):
        if value is match_value:
            return key_list[i]
    raise ValueError(f'could not find {value} in {value_list}')


def _get_module_keys_and_values_lists(module: ModuleType) -> Tuple[List[str],List[Any]]:
    key_list = []
    value_list = []
    for key, value in _module_key_value_generator(module):
        key_list.append(key)
        value_list.append(value)

    return key_list, value_list


def _module_key_value_generator(module: ModuleType) -> Iterator[Tuple[str, Any]]:
    for key in dir(module):
        yield key, getattr(module, key)


def _obj_in_module(obj, module: ModuleType) -> bool:
    for key in dir(module):
        if obj is getattr(module, key):
            return True

    return False

def _name_in_module(name: str, module: ModuleType) -> bool:
    return name in dir(module)


def _should_skip_module(name: str) -> bool:
    """
    Check if module section path ends with a section path in skip_models
    """
    module_section_path = SectionPath(name)
    for skip_name in skip_modules:
        skip_sp = SectionPath(skip_name)
        if module_section_path.endswith(skip_sp):
            return True
    return False



class Pipeline:

    def __init__(self):
        raise NotImplementedError("don't call Pipeline base class methods directly")

    def execute(self):
        raise NotImplementedError("don't call Pipeline base class methods directly")

    def new_instance_with_config(self, **kwargs):
        cls = self.__class__

        return cls(**kwargs)


from typing import List, TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.basemodels.config import ConfigBase

from pyfileconf.io.file.write.base import FileStr

from pyfileconf.imports.models.statements.container import ImportStatementContainer
from pyfileconf.assignments.models.container import AssignmentStatementContainer

class ConfigFileStr(FileStr):

    def __init__(self, config: 'ConfigBase', existing_imports: ImportStatementContainer,
                 existing_assigns: AssignmentStatementContainer, existing_body: List[str]):
        super().__init__(
            config,
            existing_imports=existing_imports,
            existing_assigns=existing_assigns,
            existing_body=existing_body
        )


from pyfileconf.assignments.logic.write import _key_value_pair_to_assignment_str

def dict_as_function_kwarg_str(d: dict) -> str:
    lines = [_key_value_pair_to_assignment_str(key, value) for key, value in d.items()]

    return '\n\t' + ',\n\t'.join(lines) + '\n'







from typing import Callable, Any
import inspect


def function_args_as_dict(func: Callable):
    args = inspect.getfullargspec(func)
    out_dict = {}

    # for methods and classmethods, want to ignore the first argument (self, cls)
    if inspect.ismethod(func):
        del args.args[0] # delete self arg

    # Handle args
    num_no_default_value = _get_length_of_arg_section(args.args) - _get_length_of_arg_section(args.defaults)
    all_defaults = [None] * num_no_default_value + _get_list_of_arg_section(args.defaults)
    assert len(all_defaults) == _get_length_of_arg_section(args.args)

    for i, arg in enumerate(args.args):
        default = all_defaults[i]
        out_dict.update({arg: default})

    # Handle kwargs
    for kwarg in args.kwonlyargs:
        if args.kwonlydefaults:
            default = args.kwonlydefaults[kwarg] if kwarg in args.kwonlydefaults else None
            out_dict.update({kwarg: default})

    return out_dict


def get_variable_name_of_obj(obj: Any) -> str:
    return [key for key, value in globals().items() if value == obj][0]


def _get_length_of_arg_section(section):
    if section is None:
        return 0

    return len(section)


def _get_list_of_arg_section(section):
    if section is None:
        return []

    return list(section)


from typing import Tuple, List

from pyfileconf.imports.models.statements.container import ImportStatementContainer
from pyfileconf.assignments.models.container import AssignmentStatementContainer

ImportsDoubleAssignsTuple = Tuple[ImportStatementContainer, AssignmentStatementContainer, AssignmentStatementContainer]

class FileStr:

    def __init__(self, import_assignment_obj, existing_imports: ImportStatementContainer,
                 existing_assigns: AssignmentStatementContainer, existing_body: List[str]):
        """

        Args:
            import_assignment_obj: object which has method obj.as_imports_and_assignments()
            existing_imports:
            existing_assigns:
            existing_body: should not contain imports, but should contain rest of existing file as str
        """
        self.file_str = self._create_str(
            import_assignment_obj=import_assignment_obj,
            existing_imports=existing_imports,
            existing_assigns=existing_assigns,
            existing_body=existing_body
        )

    def _create_str(self, import_assignment_obj, existing_imports: ImportStatementContainer,
                 existing_assigns: AssignmentStatementContainer, existing_body: List[str]) -> str:
        """
        calls obj.as_imports_and_assignments(), combines those with existing imports and
        assignments

        Args:
            import_assignment_obj:
            existing_imports:
            existing_assigns:

        Returns:

        """
        all_imports, new_assigns_begin, new_assigns_end = self._combine_imports_get_new_assignments(
            import_assignment_obj,
            existing_imports=existing_imports,
            existing_assigns=existing_assigns
        )

        # Now convert to str
        items = [
            str(all_imports) + '\n',
            str(new_assigns_begin),
            '\n'.join(existing_body),
            str(new_assigns_end)
        ]
        valid_items = [item for item in items if item not in ('', None)]
        return '\n'.join(valid_items)

    def _combine_imports_get_new_assignments(self, import_assignment_obj,
                                        existing_imports: ImportStatementContainer,
                                        existing_assigns: AssignmentStatementContainer) -> ImportsDoubleAssignsTuple:
        all_imports = existing_imports.copy()
        new_assigns_begin = AssignmentStatementContainer([])
        new_assigns_end = AssignmentStatementContainer([])

        possibly_new_imports, possibly_new_assigns = import_assignment_obj.as_imports_and_assignments()

        # Checks to see whether should be added, and whether to beginning or end, then adds
        [all_imports.add_if_missing(imp) for imp in possibly_new_imports]

        for assign in possibly_new_assigns:
            if not existing_assigns.contains_varname(assign.varname):
                begin = getattr(assign, 'prefer_beginning', False)
                if begin:
                    new_assigns_begin.append_if_missing(assign)
                else:
                    new_assigns_end.append_if_missing(assign)

        return all_imports, new_assigns_begin, new_assigns_end



from typing import Union, List, Any, TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:
    from pyfileconf.selector.models.itemview import ItemView

ListOfStrs = List[str]
StrOrListOfStrs = Union[str, ListOfStrs]
Result = Any
Results = List[Result]
ResultOrResults = Union[Result, Results]
StrOrView = Union[str, "ItemView"]
RunnerArgs = Union[str, List[str], "ItemView", List["ItemView"]]
IterativeResults = List[Tuple[Tuple[Dict[str, Any], ...], Tuple[Any]]]


import ast
import astor

def ast_node_to_source(ast_node: ast.AST) -> str:
    """
    Uses astor package to produce source code from ast

    Also handles low-level ast functions, such as wrapping in a module if necessary,
    and fixing line numbers for modified/extracted ast

    Args:
        ast_node:

    Returns:

    """

    # Must be a module to output to source. Wrap in module if not already
    if not isinstance(ast_node, ast.Module):
        ast_node = ast.Module([ast_node])

    # Fix line numbers
    ast.fix_missing_locations(ast_node)

    return astor.to_source(ast_node)


from typing import Callable, Optional, Union


def pdb_post_mortem_or_passed_debug_fn(
    *args, debug_fn: Union[bool, Callable] = False, **kwargs
) -> None:
    if isinstance(debug_fn, bool):
        if debug_fn:
            post_mortem(*args, **kwargs)
    else:
        post_mortem(*args, debug_fn=debug_fn, **kwargs)


def post_mortem(*args, debug_fn: Optional[Callable] = None, **kwargs) -> None:
    """
    Post-mortem, using a custom debug function if passed

    :param debug_fn:
    :return:
    """
    if debug_fn is None:
        import pdb

        debug_fn = pdb.post_mortem

    debug_fn()



from copy import deepcopy
from typing import Sequence, Dict, Any, List, Tuple, Type, Optional

from pyfileconf import Selector, hookimpl, IterativeRunner, reset_plugins, remove_default_plugins, PipelineManager
from pyfileconf.plugin import manager
from pyfileconf.runner.models.interfaces import RunnerArgs, ResultOrResults
from pyfileconf.sectionpath.sectionpath import SectionPath
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST

EXPECT_MODIFIED_CASES = [(dict(a=1, b=2), dict(c=3, d=4))]
EXTRA_CASE = dict(e=5, f=6)
EXTRA_RESULT = 'abc'
OVERRIDDEN_B_RESULT = 10
PRE_RUN_COUNTER = 0
POST_RUN_COUNTER = 0
PRE_UPDATE_COUNTER = 0
POST_UPDATE_COUNTER = 0


class ExamplePlugin:

    @hookimpl
    def pyfileconf_iter_get_cases(self, config_updates: Sequence[Dict[str, Any]]) -> List[Tuple[Dict[str, Any], ...]]:
        return EXPECT_MODIFIED_CASES

    @hookimpl
    def pyfileconf_iter_modify_cases(self, cases: List[Tuple[Dict[str, Any], ...]]):
        cases.append(EXTRA_CASE)

    @hookimpl
    def pyfileconf_pre_run(self, section_path_str_or_list: RunnerArgs) -> Optional[RunnerArgs]:
        global PRE_RUN_COUNTER
        PRE_RUN_COUNTER += 1
        return 'stuff.a_function'

    @hookimpl
    def pyfileconf_post_run(self, results: ResultOrResults) -> Optional[ResultOrResults]:
        global POST_RUN_COUNTER
        POST_RUN_COUNTER += 1
        return EXTRA_RESULT

    @hookimpl
    def pyfileconf_pre_update(
        self,
        pm: "PipelineManager",
        d: dict = None,
        section_path_str: str = None,
        kwargs: Dict[str, Any] = None,
    ) -> Optional[Dict[str, Any]]:
        global PRE_UPDATE_COUNTER
        PRE_UPDATE_COUNTER += 1
        return dict(b=OVERRIDDEN_B_RESULT)

    @hookimpl
    def pyfileconf_post_update(
        self,
        pm: "PipelineManager",
        d: dict = None,
        section_path_str: str = None,
        kwargs: Dict[str, Any] = None,
    ):
        global POST_UPDATE_COUNTER
        POST_UPDATE_COUNTER += 1




class PluginsTest(PipelineManagerTestBase):

    def teardown_method(self, method):
        super().teardown_method(method)
        reset_plugins()
        self.reset_counters()

    def add_plugin(self, plugin_class: Type = ExamplePlugin):
        manager.plm.register(plugin_class())
        
    def replace_plugin(self, plugin_class: Type = ExamplePlugin):
        remove_default_plugins()
        self.add_plugin(plugin_class)

    def reset_counters(self):
        global PRE_RUN_COUNTER
        global POST_RUN_COUNTER
        global PRE_UPDATE_COUNTER
        global POST_UPDATE_COUNTER
        PRE_RUN_COUNTER = 0
        POST_RUN_COUNTER = 0
        PRE_UPDATE_COUNTER = 0
        POST_UPDATE_COUNTER = 0


class TestIterPlugins(PluginsTest):

    def test_iter_no_plugins(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        cd = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=10,
            a=2
        )
        cd2 = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=20
        )
        config_dicts = [cd, cd2]
        runner = IterativeRunner(iv, config_dicts)
        assert runner.cases == [(cd,), (cd2,)]

    def test_iter_add_plugin(self):
        self.add_plugin()
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        cd = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=10,
            a=2
        )
        cd2 = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=20
        )
        config_dicts = [cd, cd2]
        runner = IterativeRunner(iv, config_dicts)
        assert runner.cases == [*EXPECT_MODIFIED_CASES, (cd,), (cd2,), EXTRA_CASE]

    def test_iter_replace_plugin(self):
        self.replace_plugin()
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        cd = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=10,
            a=2
        )
        cd2 = dict(
            section_path_str='test_pipeline_manager.stuff.a_function',
            b=20
        )
        config_dicts = [cd, cd2]
        runner = IterativeRunner(iv, config_dicts)
        assert runner.cases == [*EXPECT_MODIFIED_CASES, EXTRA_CASE]


class TestRunPlugins(PluginsTest):

    def test_run_no_plugins(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)
        result = pipeline_manager.run([iv, iv])
        assert result == [(None, None), (None, None)]
        assert POST_RUN_COUNTER == 0
        assert PRE_RUN_COUNTER == 0

    def test_run_plugins(self):
        self.add_plugin()
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == [(None, None), (None, None), EXTRA_RESULT]
        result = pipeline_manager.run([iv, iv])
        assert result == [(None, None), (None, None), (None, None), EXTRA_RESULT]
        assert POST_RUN_COUNTER == 2
        assert PRE_RUN_COUNTER == 2


class TestUpdatePlugins(PluginsTest):

    def test_update_no_plugins(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        expected_b_result = ['a', 'b']
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager.run(iv)
        assert result == (None, expected_b_result)
        assert PRE_UPDATE_COUNTER == 0
        assert POST_UPDATE_COUNTER == 0

    def test_update_with_plugins(self):
        self.add_plugin()
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        expected_b_result = ['a', 'b']
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        result = pipeline_manager.run(iv)
        assert result == [(None, OVERRIDDEN_B_RESULT), (None, OVERRIDDEN_B_RESULT), 'abc']
        assert PRE_UPDATE_COUNTER == 1
        assert POST_UPDATE_COUNTER == 1


import os
from typing import Optional, Sequence, Type, Any

from mixins.repr import ReprMixin
from pyfileconf.basemodels.collection import Collection
from pyfileconf.logic.set import _set_in_nested_obj_by_section_path
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.logic.get import _get_from_nested_obj_by_section_path
from pyfileconf.imports.models.statements.container import ImportStatementContainer

class Registrar(ReprMixin):
    repr_cols = ['name', 'basepath', 'collection']
    collection_class = Collection

    def __init__(self, collection: Collection, basepath: str, name=None,
                 always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None, klass: Optional[Type] = None,
                 key_attr: str = 'name', execute_attr: str = '__call__'):
        self.collection = collection
        self.basepath = basepath
        self.name = name

        self.always_import_strs = always_import_strs
        self.always_assign_strs = always_assign_strs
        self.klass = klass
        self.key_attr = key_attr
        self.execute_attr = execute_attr

    def __getattr__(self, item):
        return getattr(self.collection, item)

    def __dir__(self):
        exposed_methods = [
            'scaffold_config',
            'get'
        ]
        exposed_attrs = [
            'basepath',
            'name'
        ]
        collection_attrs = [attr for attr in dir(self.collection) if attr not in exposed_methods + exposed_attrs]
        return exposed_methods + exposed_attrs + collection_attrs

    @classmethod
    def from_dict(cls, dict_: dict, basepath: str, name: str=None,
                  imports: ImportStatementContainer = None, always_import_strs: Optional[Sequence[str]] = None,
                  always_assign_strs: Optional[Sequence[str]] = None, klass: Optional[Type] = None,
                  key_attr: str = 'name', execute_attr: str = '__call__'):
        collection = cls.collection_class.from_dict(
            dict_, basepath=basepath, name=name, imports=imports,
            always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
            key_attr=key_attr, execute_attr=execute_attr
        )

        return cls(
            collection, basepath=basepath, name=name,
            always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
            key_attr=key_attr, execute_attr=execute_attr
        )

    def scaffold_config(self):
        self.collection._output_config_files()

    def scaffold_config_for(self, section_path_str: str):
        section_path = SectionPath(section_path_str)

        collection = _get_from_nested_obj_by_section_path(self, section_path[:-1])
        collection._output_config_files()

    def get(self, section_path_str: str):
        section_path = SectionPath(section_path_str)

        # Goes into nested sections, until it pulls the final section or pipeline
        return _get_from_nested_obj_by_section_path(self, section_path)

    def set(self, section_path_str: str, value: Any):
        section_path = SectionPath(section_path_str)

        # Goes into nested sections, until it sets the final section or pipeline
        obj = self
        section_basepath = self.basepath
        for i, section in enumerate(section_path):
            section_basepath = os.path.join(section_basepath, section)
            try:
                obj = getattr(obj, section)
            except AttributeError as e:
                new_collection = self.collection_class(
                    section_basepath, [], name=section, imports=self.imports,
                    always_assign_strs=self.always_assign_strs,
                    always_import_strs=self.always_import_strs, klass=self.klass,
                    key_attr=self.key_attr, execute_attr=self.execute_attr
                )
                obj.append(new_collection)
                obj = getattr(obj, section)

        # Now have collection object which should hold this final object
        obj.append(value)

    def to_nested_dict(self):
        return self.collection.to_nested_dict()



from typing import Union, Any, Type

from pyfileconf.imports.logic.load.klass import class_function_args_as_dict

StrOrObj = Union[str, Any]


def convert_to_empty_obj_if_necessary(item: StrOrObj, item_class: Type, key_attr: str = 'name') -> Any:
    from pyfileconf.data.models.collection import SpecificClassCollection
    if isinstance(item, (item_class, SpecificClassCollection)):
        return item

    defaults = {
        key_attr: item
    }

    args_dict = class_function_args_as_dict(item_class)
    args_dict.update(defaults)

    return item_class(**args_dict)



from typing import Union, List, cast
import ast

AstDictOrList = Union[ast.Dict, ast.List]
DictOrList = Union[dict, list]
DictOrListOrNone = Union[DictOrList, None]


class AstDictListConverter(ast.NodeVisitor):

    def __init__(self, convert_str_values: bool = False):
        self.collections: List[DictOrList] = []
        self.convert_str_values = convert_str_values

    def visit_Dict(self, node):
        self.collections.append(
            _ast_dict_to_dict(node, convert_str_values=self.convert_str_values)
        )
        # nested nodes being handled in function above

    def visit_List(self, node):
        self.collections.append(
            _ast_list_to_list(node, convert_str_values=self.convert_str_values)
        )
        # nested nodes being handled in function above

def extract_collections_from_ast(ast_node: ast.AST, convert_str_values: bool = False) -> List[DictOrList]:
    """
    returns a list of dicts or lists. Goes through ast, converting
    ast.Dict to dict and ast.List to list, leaving the rest intact.
    Returns a list of these created dicts and lists

    Args:
        ast_node:

    Returns:

    """
    adlc = AstDictListConverter(convert_str_values=convert_str_values)
    adlc.visit(ast_node)
    return adlc.collections

def extract_collection_from_ast(ast_node: ast.AST, convert_str_values: bool = False) -> DictOrListOrNone:
    collections = extract_collections_from_ast(ast_node=ast_node, convert_str_values=convert_str_values)

    if len(collections) == 0:
        return None

    if len(collections) > 1:
        raise ValueError(f'expected to extract one assignment from ast. got {len(collections)} '
                         f'assigns: {collections}')

    return collections[0]


def _ast_dict_or_list_to_dict_or_list(node: AstDictOrList, convert_str_values: bool = False) -> DictOrList:
    if isinstance(node, ast.Dict):
        return _ast_dict_to_dict(node, convert_str_values=convert_str_values)
    elif isinstance(node, ast.List):
        return _ast_list_to_list(node, convert_str_values=convert_str_values)
    else:
        raise ValueError(f'expected ast.Dict or ast.List. Got {node} of type {type(node)}')


def _ast_dict_to_dict(ast_dict: ast.Dict, convert_str_values: bool = False) -> dict:
    out_dict = {}
    for key, value in zip(ast_dict.keys, ast_dict.values):
        key = cast(ast.Str, key)
        key_string = key.s
        if isinstance(value, (ast.Dict, ast.List)):
            store_value = _ast_dict_or_list_to_dict_or_list(value, convert_str_values=convert_str_values)
        else:
            store_value = _convert_to_str_if_ast_str_and_desired(value, convert_desired=convert_str_values)
        out_dict[key_string] = store_value

    return out_dict


def _ast_list_to_list(ast_list: ast.List, convert_str_values: bool = False) -> list:
    out_list = []
    for item in ast_list.elts:
        if isinstance(item, (ast.Dict, ast.List)):
            store_item = _ast_dict_or_list_to_dict_or_list(item, convert_str_values=convert_str_values)
        else:
            store_item = _convert_to_str_if_ast_str_and_desired(item, convert_desired=convert_str_values)
        out_list.append(store_item)

    return out_list

def _convert_to_str_if_ast_str_and_desired(ast_node: ast.AST, convert_desired=False):
    if not convert_desired:
        return ast_node

    if isinstance(ast_node, ast.Str):
        return ast_node.s

    return ast_node


from typing import Optional

from pyfileconf.sectionpath.sectionpath import SectionPath


class RunningTracker:
    orig_running_section_path_str: Optional[str]

    def __init__(self, section_path_str: str, base_section_path_str: Optional[str] = None):
        if base_section_path_str is not None:
            new_sp = SectionPath.join(base_section_path_str, section_path_str)
            section_path_str = new_sp.path_str
        self.section_path_str = section_path_str
        self.orig_running_section_path_str = None

    def __enter__(self):
        from pyfileconf import context
        self.orig_running_section_path_str = context.currently_running_section_path_str
        context.currently_running_section_path_str = self.section_path_str

    def __exit__(self, exc_type, exc_val, exc_tb):
        from pyfileconf import context
        context.currently_running_section_path_str = self.orig_running_section_path_str
